/*  Exile - IRC to Relational Database Gateway
 *  Copyright (C) 2002 Lucas Nussbaum <lucas@lucas-nussbaum.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: actions.c 2215 2008-08-04 23:48:13Z sirvulcan $
 */

#include "exile.h"
#include "actions.h"
#include "db.h"
#include "hosthiding.h"
#include "send.h"
#include "misc.h"
#include "compat.h"
#include "memory.h"
#include "log.h"
#include "sockutil.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern char *BotNick;
extern char *ServerNum;
extern char *ServerName;
extern char *RemoteServer;
extern time_t start_time;
extern char inbuf[];
extern int ServerCacheTime;
extern int UserCacheTime;
extern int HidePasswords;
extern int NefariousHostCrypt;
extern int localclients;
extern int eob_ack;
extern int servsock;
extern long query_count;

int last_stats_l;
int stats_l_sent;

MYSQL *myptr;

/*************************************************************************/
/*
 * Max values calculation
 */

unsigned int nbusers = 0;
unsigned int nbusers_max = 0;
unsigned int nbchans = 0;
unsigned int nbchans_max = 0;
unsigned int nbservs = 0;
unsigned int nbservs_max = 0;
/* keep track of logged umodes et cmodes, init to 0 by default */
int log_umode[256];
int log_cmode[256];

/* check if nbchans > nbchans_max */
void do_checknbchansmax()
{
	if (nbchans > nbchans_max)
	{
		nbchans_max = nbchans;
		db_query
			("UPDATE " TBL_MAXV
			 " SET val='%d', time=NOW() WHERE type='channels'",
			 nbchans_max);
	}
}

/* check if nbusers > nbusers_max */
void do_checknbusersmax()
{
	if (nbusers > nbusers_max)
	{
		nbusers_max = nbusers;
		db_query
			("UPDATE " TBL_MAXV
			 " SET val='%d', time=NOW() WHERE type='users'", nbusers_max);
	}
}

/* check if nbservs > nbservs_max */
void do_checknbservsmax()
{
	if (nbservs > nbservs_max)
	{
		nbservs_max = nbservs;
		db_query
			("UPDATE " TBL_MAXV
			 " SET val='%d', time=NOW() WHERE type='servers'", nbservs_max);
	}
}

/*************************************************************************/
/*
 * general purpose functions
 */

void do_stats()
{
	char **res;
	MYSQL_RES *resptr;
	int rows = 0;

	/* STATS l */
	db_query("SELECT COUNT(*) FROM " TBL_STATS_l);
        resptr = mysql_store_result(myptr);
        if (mysql_num_rows(resptr))
                rows = atoi(*mysql_fetch_row(resptr));
        mysql_free_result(resptr);

        /* oh god ill explain this if 
         * First part is we only do a stats l every 900 seconds (15mins) OR
         * (Second Part) if there are no rows in exile_stats_l and it has been
	 * atleast 30 seconds since we finished the burst then we do a stats l
	 * to fill the table. In both cases dont do the stats l if there has
	 * been one sent (stats_l_sent == 1) */
        if (((last_stats_l + 900) < time(NULL) && (last_stats_l > 900) && BotNick && (stats_l_sent == 0)) || 
            ((stats_l_sent == 0) && (rows == 0) && (eob_ack > 30) && ((eob_ack + 30) < time(NULL)))) {
		db_query("SELECT snumeric FROM " TBL_SERV);

        	resptr = mysql_store_result(myptr);
	        while ((res = mysql_fetch_row(resptr)))
	        {
			char *numeric = db_escape(res[0]);

			send_cmd(NULL, "%sAAA R l :%s", ServerNum, numeric);
		}
		stats_l_sent = 1;
		last_stats_l = time(NULL);
	}
}

/* adds modes to a given chanid */
static void do_chanmodes(int chanid, char **av, int ac, int clearmodes)
{
	/* the ircd parses the mode before forwarding it to exile, so there's no buffer overrun
	 * possibility here.   - lucas
	 */
	int atleastone = 0;
	char db[1000];					  /* should be enough for long queries */
	char tmp[14] = "mode_XX=\"X\", ";
	char *modes = av[0];
	int argptr = 1;
	if (*modes == '0')
		return;
	strcpy(db, "UPDATE " TBL_CHAN " SET ");
	while (*modes)
	{
		switch (*modes)
		{
		case '+':
			tmp[9] = 'Y';
			break;
		case '-':
			tmp[9] = 'N';
			break;
		default:
			if (clearmodes == 1)
				tmp[9] = 'N';
			if (!strchr(CHANMODES, *modes))
			{
				if (!log_cmode[(int) *modes])
				{
					mylog("unknown mode : chanmode %c (in %s)", *modes, inbuf);
					log_cmode[(int) *modes] = TRUE;
				}
			}
			else if (*modes == 'I')	/* ignore them */
				argptr++;

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
			else if (*modes == 'o' || *modes == 'v' || *modes == 'h')
#else
			else if (*modes == 'o' || *modes == 'v')
#endif
			{
				char *user;
				user = db_escape(av[argptr++]);
				db_query
					("UPDATE " TBL_ISON
					 " SET mode_l%c='%c' WHERE chanid='%d' AND nickid='%d'",
					 *modes, tmp[9], chanid, db_getnumeric(user));
			}
			else
			{
				atleastone = 1;
				tmp[5] = ((*modes >= 'a') ? 'l' : 'u');
				tmp[6] = tolower(*modes);
				strcat(db, tmp);
				if (*modes == 'b') {
					char *ban = db_escape(av[argptr++]);

					if (tmp[9] == 'Y')
					{
						db_query("INSERT INTO " TBL_CHBAN
							" (chanid, ban) VALUES ('%d', '%s')",
							chanid, ban);
					} else {
						db_query("DELETE FROM " TBL_CHBAN " WHERE chanid='%d' AND ban='%s'", chanid, ban);
					}
				}
				#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
				else if (*modes == 'e') {
					char *exempt = db_escape(av[argptr++]);

					if (tmp[9] == 'Y')
					{
						db_query("INSERT INTO " TBL_CHEXEMPT
							" (chanid, exempt) VALUES ('%d', '%s')",
							chanid, exempt);
					} else {
						db_query("DELETE FROM " TBL_CHEXEMPT " WHERE chanid='%d' AND exempt='%s'", chanid, exempt);
					}
				}
				#endif				
				else if (*modes == 'k')
				{
					if (tmp[9] == 'Y')
					{
						char *key = db_escape(av[argptr++]);
						if (HidePasswords)
						{
							strcat(db, "mode_lk_data='HIDDEN', ");
						}
						else
						{
							strcat(db, "mode_lk_data='");
							strcat(db, key);
							strcat(db, "', ");
						}
					}
					else
					{
						strcat(db, "mode_lk_data='', ");
						argptr++;	  /* mode -k needs a parameter */
					}
				}
				#if defined(IRCD_IRCU12)
				else if (*modes == 'A')
				{
					if (tmp[9] == 'Y')
					{
						char *key = db_escape(av[argptr++]);
						if (HidePasswords)
						{
							strcat(db, "mode_uA_data='HIDDEN', ");
						}
						else
						{
							strcat(db, "mode_uA_data='");
							strcat(db, key);
							strcat(db, "', ");
						}
					}
					else
					{
						strcat(db, "mode_uA_data='', ");
						argptr++;	  /* mode -A needs a parameter */
					}
				}
				else if (*modes == 'U')
				{
					if (tmp[9] == 'Y')
					{
						char *key = db_escape(av[argptr++]);
						if (HidePasswords)
						{
							strcat(db, "mode_uU_data='HIDDEN', ");
						}
						else
						{
							strcat(db, "mode_uU_data='");
							strcat(db, key);
							strcat(db, "', ");
						}
					}
					else
					{
						strcat(db, "mode_uU_data='', ");
						argptr++;	  /* mode -U needs a parameter */
					}
				}
				#endif
				else if (*modes == 'l')
				{
					if (tmp[9] == 'Y')
					{
						strcat(db, "mode_ll_data='");
						strcat(db, av[argptr++]);
						strcat(db, "', ");
					}
					else
					{
						strcat(db, "mode_ll_data='', ");
					}
				}
			}
			break;
		}
		modes++;
	}
#ifndef NOMODES
	if (atleastone)
	{
		sprintf(&db[strlen(db) - 2], " WHERE chanid='%d'", chanid);
		db_query(db);
	}
#endif /* !NOMODES */
}

/* add one or more users to a chanid */
void do_addusers(int chanid, char *users, int create)
{
	int op, voice, halfop, owner, protect;
	char *nextusers;
	int nickid;
	while (users && (*users))
	{
		/* Neostats is broken and sends an additional space. */
		if (*users == ' ')
			users++;
		if (create == 1)
			op = 1;
		else
			op = 0;
		halfop = 0;
		voice = 0;
		owner = 0;
		protect = 0;

		if (*users == '@')
		{
			users++;
			op = 1;
		}
		if (*users == 'h')
		{
			users++;
			halfop = 1;
		}
		if (*users == '+')
		{
			users++;
			voice = 1;
		}

		nextusers = strchr(users, ' ');
		if (nextusers)
			*nextusers = '\0';
		users = db_escape(users);
		nickid = db_checknumeric(users);
		if (nickid != -1)
		{
#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
                        MYSQL_RES *resptr;
			int rcount = 0;
			db_query("INSERT INTO " TBL_ISON
						" (nickid, chanid, mode_lo, mode_lh, mode_lv) VALUES ('%d', '%d','%s','%s','%s')",
						nickid, chanid, (op ? "Y" : "N"), (halfop ? "Y" : "N"), (voice ? "Y" : "N"));

			db_query("SELECT count(*) FROM " TBL_ISON " where chanid='%d'", chanid);
			resptr = mysql_store_result(myptr);
			if (mysql_num_rows(resptr)) {
				rcount = atoi(*mysql_fetch_row(resptr));
				mylog("debug: actions.c adduser");
                                db_query("UPDATE " TBL_CHAN " SET usercount='%d' where chanid='%d'", rcount, chanid);
			}
                        mysql_free_result(resptr);
#else
			db_query("INSERT INTO " TBL_ISON
						" (nickid, chanid, mode_lo, mode_lv) VALUES ('%d', '%d','%s','%s')",
						nickid, chanid, (op ? "Y" : "N"), (voice ? "Y" : "N"));
#endif
		}
		else
		{
			mylog("received join of non-existing user %s on channel ID %d",
					users, chanid);
		}
		users = nextusers;
		if (users)
			users++;
	}
}

/* add modes to a nickid */
void do_usermodes(int nickid, char *modes, char **args, int nbmodes)
{
#ifndef NOMODES
	int atleastone = 0;
	char db[1000];					  /*should be enough for long queries */
	char tmp[14] = "mode_XX='X', ";
	char *uaccount = "";
	char *useraccount = "";
	char *timestamp = "";
	char *accb, *acc = NULL;
        int p = 5, ar = 2, c = 1;
	strcpy(db, "UPDATE " TBL_USER " SET ");

	while (*modes)
	{
		switch (*modes)
		{
		case '+':
			tmp[9] = 'Y';
			ar = 1;
			break;
		case '-':
			tmp[9] = 'N';
			ar = 0;
			break;
		default:
#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
			switch (*modes) {
				case 'r':
					if (ar == 1) {
						p++;
						uaccount = db_escape(args[p]);
					        for (acc = strtok_r(uaccount, ":", &accb);
					                acc;
					                acc = strtok_r(NULL, ":", &accb)) {
					                if (c == 1)
					                        useraccount = acc;
					                else if (c == 2)
					                        timestamp = acc;

					                c++;
					        }

					        db_query("UPDATE " TBL_USER
					                 " SET account='%s' WHERE nickid='%d'",
			        		         useraccount, nickid);
					        db_query("UPDATE " TBL_USER
					                 " SET acctimestamp='%s' WHERE nickid='%d'",
			        		         timestamp, nickid);
					}
					break;
				case 'h':
					if (ar == 1) {
						p++;
					        db_query("UPDATE " TBL_USER
					                 " SET sethost='%s' WHERE nickid='%d'",
			        		         db_escape(args[p]), nickid);
					}
					break;
				case 'f':
					if (ar == 1) {
						p++;
					        db_query("UPDATE " TBL_USER
					                 " SET fakehost='%s' WHERE nickid='%d'",
			        		         db_escape(args[p]), nickid);
					}
					break;
                                case 'C':
                                        if (ar == 1) {
						p++;
                                                db_query("UPDATE " TBL_USER
                                                         " SET virthost='%s' WHERE nickid='%d'",
                                                         db_escape(args[p]), nickid);
                                        }
                                        break;
                                case 'c':
                                        if (ar == 1) {
						p++;
                                                db_query("UPDATE " TBL_USER
                                                         " SET virtip='%s' WHERE nickid='%d'",
                                                         db_escape(args[p]), nickid);
                                        }
                                        break;
			}
#endif
			if (!strchr(USERMODES, *modes))
			{
				if (!log_umode[(int) *modes])
				{
					mylog("unknown mode : usermode %c (in %s)", *modes, inbuf);
					log_umode[(int) *modes] = TRUE;
				}
			}
			else
			{
				atleastone = 1;
				tmp[5] = ((*modes >= 'a') ? 'l' : 'u');
				tmp[6] = tolower(*modes);
				strcat(db, tmp);
			}
			break;
		}
		modes++;
	}
	if (atleastone)
	{
		sprintf(&db[strlen(db) - 2], " WHERE nickid='%d'", nickid);
		db_query(db);
	}
#endif /* !NOMODES */
}

/*************************************************************************/
/*
 * IRC messages handlers
 */

/* SERVER */
void do_server(char *server, char *comment, char *linkedto, char *numeric)
{
	/* Aa S gnuworld.afternet.org 2 0          1122878038 J10 Az]]] +s :Services */
      /* SERVER afro4.funkpolice.com  1 1121933417 1123010310 J10 AaAP] +h :Hell IRC Network */
	int servid;
	int add = 1;
	char num[2];
	strcpy(num, numeric);

	num[2] = '\0';
	numeric = num;
	server = db_escape(server);
	comment = db_escape(comment);
	linkedto = db_escape(linkedto);
	if (ServerCacheTime && ((servid = db_checkserver(server)) != -1))
	{
		db_query("UPDATE " TBL_SERV
					" SET server='%s', snumeric='%s', comment='%s', linkedto='%d', connecttime=NOW(), online='Y' WHERE servid='%d'",
					server, num, comment, db_getserver(linkedto), servid);
		db_addserver(server, servid);
		db_addserver(num, servid);
		add = 0;
	}
	if (add)
	{
		db_query
			("INSERT INTO " TBL_SERV
			 " VALUES('', '%s', '%s', '%s', '%d', NOW(), 'Y', '')",
			 server, num, comment, db_getserver(linkedto));
		db_addserver(server, db_insertid());
		db_addserver(num, db_insertid());
	}
	if (ServerCacheTime)
		db_cleanserver();
	nbservs++;
	do_checknbservsmax();
}


void servercleanup(int servid, int cached)
{
	int nickid;
	char *snick, *numeric;
	char **res;
	char **res2;
	MYSQL_RES *resptr;
	MYSQL_RES *resptr2;

	db_query("SELECT servid,server,snumeric FROM " TBL_SERV " WHERE linkedto='%d'", servid);

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
        {
       	        int oservid = atoi(db_escape(res[0]));
       	        char *oserver = db_escape(res[1]);
		char *snumeric = db_escape(res[2]);


		db_query("SELECT nickid,nick,snumeric FROM " TBL_USER " WHERE servid='%d'", oservid);

	        resptr2 = mysql_store_result(myptr);
	        while ((res2 = mysql_fetch_row(resptr2)))
	        {
	                nickid = atoi(db_escape(res2[0]));
			snick = res2[1];
			numeric = res2[2];

			db_removenickid(nickid);
			if (!UserCacheTime) {
                	        db_delnumeric(numeric);
	                	db_delnick(snick);  
				unset_geoip_info(nickid);
			}
	        }
		mysql_free_result(resptr2);

		servercleanup(oservid, cached);
		if (cached) {
			db_query("UPDATE " TBL_SERV
					" SET online='N', lastsplit=NOW(),linkedto=NULL WHERE server='%s'",
					oserver);
		} else {
			db_query("DELETE FROM " TBL_SERV " WHERE server='%s'", oserver);
			db_delserver(oserver);
			db_delserver(snumeric);
		}
	}
	mysql_free_result(resptr);
}

void check_me(int servid, char *reason)
{
	int id, myid, rc = 0;
	char **res;
	char *server = NULL;
	MYSQL_RES *resptr;

	db_query("SELECT servid FROM " TBL_SERV " WHERE server='%s'", ServerName);
	resptr = mysql_store_result(myptr);
	while ((res = mysql_fetch_row(resptr)))
		myid = atoi(res[0]);

	db_query("SELECT servid,server FROM " TBL_SERV " WHERE linkedto='%d'", myid);
	resptr = mysql_store_result(myptr);
	while ((res = mysql_fetch_row(resptr))) {
		id = atoi(res[0]);
		server = res[1];

		if (id == servid) {
			if (!(strcmp(reason, "Ping timeout")))
				rc = 1;
			else if (!(strcmp(reason, "<>")))
				rc = 1;

			if (rc == 0) {
				disconn(servsock);
				exit(1);
			}
		}
	}
}

/* SQUIT */
void do_squit(char *server, char *reason)
{
	int nickid;
	int servid;
	char **res;
	char *snick;
	char *snumeric;
	char *hostname;
	MYSQL_RES *resptr;

	server = db_escape(server);
	servid = db_getserver(server);

	check_me(servid, reason);

	db_query("SELECT nickid, nick, hostname, snumeric FROM " TBL_USER " WHERE servid='%d'", servid);

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
        {
                nickid = atoi(res[0]);
                snick = res[1];
		hostname = res[2];
                snumeric = res[3];

		db_removenickid(nickid);
		if (!UserCacheTime) {
                        db_delnumeric(snumeric);
	                db_delnick(snick);  
			unset_geoip_info(nickid);
		}
        }
        mysql_free_result(resptr);


	if (ServerCacheTime)
	{
		servercleanup(servid, 1);

		db_query("UPDATE " TBL_SERV
					" SET online='N', lastsplit=NOW(),linkedto=NULL WHERE servid='%d'",
					db_getserver(server));
		db_cleanserver();
	}
	else
	{
		servercleanup(servid, 0);

		db_query("SELECT snumeric FROM " TBL_SERV " WHERE servid='%d'", servid);
	        resptr = mysql_store_result(myptr);
	        while ((res = mysql_fetch_row(resptr)))
	        {
			char *snumeric = db_escape(res[0]);
			db_delserver(snumeric);
	        }
		mysql_free_result(resptr);

		db_query("DELETE FROM " TBL_SERV " WHERE server='%s'", server);
		db_delserver(server);
	}

        db_query("DELETE FROM " TBL_STATS_l " WHERE servid='%d'", server);

	if (UserCacheTime)
		db_cleanuser();

	nbservs--;
}

/* NICK (new nick) */
void do_nick_new(int ac, char **av, char *source)
{
	char *nick = 0;
	char *realname = 0;
	char *hostname = 0;
	char *username = 0;
	char *serv = 0;
	char *numeric = 0;
	char *oldsnumeric = 0;
	char *uaccount = "";
	char *useraccount = "";
	char *ipaddr;
        char *accb, *acc = NULL;
	char *timestamp = ""; 
	char **res;
	MYSQL_RES *resptr;

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
	int fakehost = 0;
	int sethost = 0;
        int virthost = 0;
        int virtip = 0;
	char *usethost = "";
	char *ufakehost = "";
	char *uvirthost = "";
	char *uvirtip = "";
#endif

	unsigned int connecttime;
	int servid;
	int add = 1;
	int nickid = 0;
	int modes = 0;
	int account = 0;
	int c = 1;
	unsigned int IP;
	char* dq;

	nick = db_escape(av[0]);
	connecttime = atoi(av[2]);
	username = db_escape(av[3]);
	hostname = db_escape(av[4]);

/* AI N Flash 2 1217299573 Flash Scream.NL.AfterNET.Org +oiwgrhf flash:1072391267 Flash@afternet.services flash.Staff.Afte$ */
	if (strpbrk(av[5], "+"))
		modes = 1;
	realname = db_escape(av[ac-1]);
	numeric =  db_escape(av[ac-2]);
	IP =       base64toint(av[ac-3]);
        dq = unsigned32ToDotQuad(IP);
	ipaddr = dq;

	serv = source;
	if (modes == 1) {
		int p = 6;

		if (strchr(av[5], 'r')) {
			account = 1;
		}
		if (strchr(av[5], 'f')) {
			fakehost = 1;
		}
		if (strchr(av[5], 'h')) {
			sethost = 1;
		}
		if (strchr(av[5], 'C')) {
			virthost = 1;
		}
		if (strchr(av[5], 'c')) {
			virthost = 1;
		}

		p = 5;
		if (account == 1) {
			p++;
			uaccount = db_escape(av[p]);
		}

		if (sethost == 1) {
			p++;
			usethost = db_escape(av[p++]);
		}

		if (fakehost == 1) {
			p++;
			ufakehost = db_escape(av[p++]);
		}

		if (virthost == 1) {
			p++;
			uvirthost = db_escape(av[p++]);
		}

		if (virtip == 1) {
			p++;
			uvirtip = db_escape(av[p++]);
		}
	}

	for (acc = strtok_r(uaccount, ":", &accb);
		acc;
		acc = strtok_r(NULL, ":", &accb)) {
		if (c == 1)
			useraccount = acc;
		else if (c == 2)
			timestamp = acc;
			
		c++;
	}
	c = 1;

	servid = db_getserver(serv);
	if (UserCacheTime && ((nickid = db_checknick(nick)) != -1))
	{
		db_query("SELECT snumeric FROM " TBL_USER " WHERE nickid='%d'", nickid);
	        resptr = mysql_store_result(myptr);
		while ((res = mysql_fetch_row(resptr)))
			oldsnumeric = db_escape(res[0]);

	        mysql_free_result(resptr);

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
		db_query
			("UPDATE " TBL_USER
			 " SET nick='%s', snumeric='%s', account='%s', acctimestamp='%s', realname='%s', hostname='%s', sethost='%s', fakehost='%s', virthost='%s', virthost='%s', ipaddr='%s', username='%s', connecttime=FROM_UNIXTIME('%d'), servid='%d',lastquit=NULL, online='Y', away='N', awaymsg='' WHERE nickid='%d'",
			 nick, numeric, useraccount, timestamp, realname, hostname, usethost, ufakehost, uvirthost, uvirtip, ipaddr, username, connecttime, servid,
#else
		db_query
			("UPDATE " TBL_USER
			 " SET nick='%s', snumeric='%s', account='%s', acctimestamp='%s', realname='%s', hostname='%s', ipaddr='%s', username='%s', connecttime=FROM_UNIXTIME('%d'), servid='%d',lastquit=NULL, online='Y', away='N', awaymsg='' WHERE nickid='%d'",
			 nick, numeric, useraccount, timestamp, realname, hostname, ipaddr, username, connecttime, servid,
#endif
			 nickid);

		db_updatehash(numeric, oldsnumeric, nickid);
		add = 0;
	}
	if (add)
	{
#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
		db_query
			("INSERT INTO " TBL_USER
			 " (nick, snumeric, account, acctimestamp, realname, hostname, sethost, fakehost, virthost, virtip, ipaddr, username, connecttime, servid) VALUES('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', FROM_UNIXTIME('%d'), '%d')",
			 nick, numeric, useraccount, timestamp, realname, hostname, usethost, ufakehost, uvirthost, uvirtip, ipaddr, username, connecttime, servid);
#else
		db_query
			("INSERT INTO " TBL_USER
			 " (nick, snumeric, account, acctimestamp, realname, hostname, ipaddr, username, connecttime, servid) VALUES('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', FROM_UNIXTIME('%d'), '%d')",
			 nick, numeric, useraccount, timestamp, realname, hostname, ipaddr, username, connecttime, servid);
#endif
		db_addnick(nick, db_insertid());
		db_addnumeric(numeric, db_insertid());

	}

	nickid = db_checknick(nick);
	set_geoip_info(nickid, hostname);

	if (UserCacheTime)
		db_cleanuser();

	if (modes == 1)
		do_usermodes(nickid, av[5], av, ac);
	nbusers++;
	do_checknbusersmax();

}

/* NICK (nick change) */
void do_nick_chg(char *newnick, char *oldnick)
{
	int nickid;
	int backupid;
        char **res;
        MYSQL_RES *resptr;

	oldnick = db_escape(oldnick);
	newnick = db_escape(newnick);

	nickid = db_getnumeric(oldnick);
	backupid = nickid;

        db_query("SELECT nick FROM " TBL_USER " WHERE nickid='%d'", nickid);

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
                oldnick = db_escape(res[0]);

        mysql_free_result(resptr);

	/* the target nickname might already exist if caching is enabled */
	if (UserCacheTime && (strcasecmp(newnick, oldnick))
		 && ((nickid = db_checknick(newnick)) != -1))
	{
		/* In this case, we don't keep a record of the old nick. It would be :
		 * - technically difficult, because we'd have to make a copy of the record
		 * - dangerous, because it would provide an easy way to fill up the DB */
		db_delnick(newnick);
		db_query("DELETE from " TBL_USER " WHERE nickid='%d'", nickid);
	}
	/* we update the nickname */
	db_query("UPDATE " TBL_USER " SET nick='%s' WHERE nickid='%d'",
				newnick, backupid);
	db_chgnick(newnick, oldnick);
}

/* PART */
void do_part(char *chan, char *nick)
{
	int nickid, chanid;
	MYSQL_RES *resptr;
	int rcount = 0;

        clean_channelname(chan);
	chan = db_escape(chan);

	nick = db_escape(nick);
        nickid = db_getnumeric(nick);
        chanid = db_getchannel(chan);

	db_query("DELETE FROM " TBL_ISON
				" WHERE nickid='%d' AND chanid='%d'", nickid, chanid);


	db_query("SELECT count(*) FROM " TBL_ISON " where chanid='%d'", chanid);
	resptr = mysql_store_result(myptr);
	if (mysql_num_rows(resptr)) {
		rcount = atoi(*mysql_fetch_row(resptr));
		mylog("debug: actions.c: part");
		db_query("UPDATE " TBL_CHAN " SET usercount='%d' where chanid='%d'", rcount, chanid);
	}
	mysql_free_result(resptr);

	db_checkemptychan(db_getchannel(chan), chan);
}

/* JOIN - join 0 */
void do_partall(char *nick)
{
	nick = db_escape(nick);
	db_removefromchans(db_getnumeric(nick));
}

/* JOIN */
void do_join(char *chan, char *nick, int create)
{
	/* we don't escape nick because do_addusers does. */
	do_addusers(db_getchancreate(chan), nick, create);
}

/* QUIT */
void do_quit(char *nick)
{
	char *num = strdup(nick);
	char *snum;
	char *snumeric = NULL;
	char *snick = NULL;
	int nickid;

	snum = db_escape(num);

	if (UserCacheTime == 0) {
		char **res;
		MYSQL_RES *resptr;

		db_query("SELECT nickid,nick,snumeric FROM " TBL_USER " WHERE snumeric='%s'", snum);

        	resptr = mysql_store_result(myptr);
	        while ((res = mysql_fetch_row(resptr))) {
			nickid = atoi(res[0]);
        		snick = db_escape(res[1]);
        		snumeric = db_escape(res[2]);
		}

		mysql_free_result(resptr);
	}

	db_removenick(snum);			  /* sql */
	if (UserCacheTime)
	{
		db_cleanuser();
	}
	else
	{
		if (snumeric)
			db_delnumeric(snum);
		db_delnick(snick);			  /* hash */
		unset_geoip_info(nickid);
	}

	db_query("DELETE FROM " TBL_SILENCE " WHERE nickid='%d'", nickid);

	nbusers--;
}

/* [SVS]MODE - user */
void do_umode(char *nick, char *modes, char **args, int nbmodes)
{
	nick = db_escape(nick);
	do_usermodes(db_getnick(nick), modes, args, nbmodes);
}

/* [SVS]MODE - channel */
void do_cmode(char *chan, char **modes, int nbmodes)
{
        clean_channelname(chan);
	chan = db_escape(chan);

	do_chanmodes(db_getchannel(chan), modes, nbmodes, 0);
}

void do_clearmode(char *chan, char **modes, int nbmodes)
{
        clean_channelname(chan);
	chan = db_escape(chan);

	do_chanmodes(db_getchannel(chan), modes, nbmodes, 1);
}

/*
 * Remove bells and commas from channel name
 */
void clean_channelname(char *cn)
{
  int i;

  for (i = 0; cn[i]; i++) {
    if (IsChannelLower(cn[i])) {
      cn[i] = ToLower(cn[i]);

      if ((unsigned char)(cn[i]) == 0xd0)
        cn[i] = (char) 0xf0;
    }
  }
}

void do_topic(char *chan, char *author, char *time, char *topic)
{
        clean_channelname(chan);
	chan = db_escape(chan);

	author = db_escape(author);
	topic = db_escape(topic);
	db_query
		("UPDATE " TBL_CHAN
		 " SET topic='%s', topicauthor='%s', topictime=FROM_UNIXTIME('%s') WHERE chanid='%d'",
		 topic ? topic : "", author, time, db_getchannel(chan));
}

/* AWAY set */
void do_away_set(char *nick, char *msg)
{
	nick = db_escape(nick);
	msg = db_escape(msg);
	db_query("UPDATE " TBL_USER
				" SET away='Y', awaymsg='%s' WHERE nickid='%d'", msg,
				db_getnumeric(nick));
}

/* AWAY unset */
void do_away_unset(char *nick)
{
	nick = db_escape(nick);
	db_query("UPDATE " TBL_USER
				" SET away='N', awaymsg='' WHERE nickid='%d'",
				db_getnumeric(nick));
}

/* LUSERS queries */
void do_lusers(char *nick)
{
	send_cmd(NULL,
				"%s 251 %s :There are %d users and %d invisible on %d servers",
				ServerNum, nick, db_getlusers(LUSERS_USERS),
				db_getlusers(LUSERS_USERSINV), db_getlusers(LUSERS_SERV));
	send_cmd(NULL,		"%s 252 %s :%d :operator(s) online", ServerNum, nick,
				db_getlusers(LUSERS_OPERS));
	send_cmd(NULL, 		"%s 254 %s :%d :channels formed", ServerNum, nick, nbchans,
				db_getlusers(LUSERS_CHAN));
	send_cmd(NULL, 		"%s 255 %s :I have %d clients and 1 servers", ServerNum, nick, localclients);
	send_cmd(NULL, 		"%s 265 %s :Current local users: %d Max: %d", ServerNum, nick, localclients, localclients);
	send_cmd(NULL, 		"%s 266 %s :Current global users: %d Max: %d", ServerNum, nick,
				db_getlusers(LUSERS_USERSGLOB), db_getlusers(LUSERS_USERSMAX));

	send_cmd(NULL, "%s O %s :SQL Query Count: %d", ServerNum, nick, query_count);
}

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
#define SWHOISBUGFREQ 20

void do_swhois(char *user, char *msg)
{
	int nickid;
	user = db_escape(user);
	msg = db_escape(msg);
	nickid = db_getnumeric_unsure(user);

	db_query
		("UPDATE " TBL_USER " SET swhois='%s' WHERE nickid='%d'",
		 msg, nickid);

}

void do_silence(char *user, char *host)
{
	char *dhost = NULL;
	int nickid;
	int remove = 0;
	char **res;
	MYSQL_RES *resptr;

	user = db_escape(user);

	nickid = db_getnumeric_unsure(user);
	
	if (*host == '-')
		remove = 1;

	host = host + 1;

	db_query("SELECT nickid,host FROM " TBL_SILENCE " WHERE nickid='%d' AND host='%s'", nickid,host);

       	resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
		dhost = res[1];

	mysql_free_result(resptr);

	if (dhost && (remove == 1))
		db_query("DELETE FROM " TBL_SILENCE " WHERE nickid='%d' AND host='%s'", nickid, host);		
	else if (!dhost && (remove == 0))
		db_query("INSERT INTO " TBL_SILENCE " (nickid, host) VALUES ('%d', '%s')", nickid, host);

}
#endif
